from datetime import datetime
import os
from typing import Optional, Union

from anylearn.interfaces.base import BaseObject
from anylearn.interfaces.resource.resource_uploader import ResourceUploader
from anylearn.interfaces.resource.resource_downloader import ResourceDownloader
from anylearn.utils.errors import AnyLearnException
from anylearn.utils.api import (
    url_base,
    get_with_token,
    post_with_token,
)


class Resource(BaseObject):
    """
    Attributes
    ----------
    id
        资源的唯一标识符，自动生成，由类名前四位+uuid1生成的编码中后28个有效位（小写字母和数字）组成（自动生成）
    name
        资源的名称（长度1~50）
    description
        资源描述（长度最大255）
    state
        资源状态
    visibility
        资源可见性（默认3）
    upload_time
        资源上传时间
    filename
        下一步中会被分片上传的资源的完整名（包括扩展名）
    is_zipfile
        是否为zip文件
    file_path
        资源文件路径
    size
        资源文件大小
    creator_id
        创建者ID
    node_id
        节点ID
    owner
        资源的所有者，以逗号分隔的这些用户的ID拼成的字符串，无多余空格
    load_detail
        初始化时是否加载详情
    """

    """具体资源信息配置"""
    _fields = {
        # 资源创建/更新请求包体中必须包含且不能为空的字段
        'required': {
            'create': ['name', 'filename'],
            'update': ['id', 'name'],
        },
        # 资源创建/更新请求包体中包含的所有字段
        'payload': {
            'create': ['name', 'description', 'visibility', 'owner',
                       'filename'],
            'update': ['id', 'name', 'description', 'visibility', 'owner'],
        },
    }

    def __init__(self,
                 id: Optional[str]=None,
                 name: Optional[str]=None,
                 description: Optional[str]=None,
                 state: Optional[int]=None,
                 visibility=3,
                 upload_time: Optional[Union[datetime, str]]=None,
                 filename: Optional[str]=None,
                 is_zipfile: Optional[int]=None,
                 file_path: Optional[str]=None,
                 size: Optional[str]=None,
                 creator_id: Optional[str]=None,
                 node_id: Optional[str]=None,
                 owner: Optional[list]=None,
                 load_detail=False):
        """
        Parameters
        ----------
        id
            资源的唯一标识符，自动生成，由类名前四位+uuid1生成的编码中后28个有效位（小写字母和数字）组成（自动生成）
        name
            资源的名称（长度1~50）
        description
            资源描述（长度最大255）
        state
            资源状态
        visibility
            资源可见性（默认3）
        upload_time
            资源上传时间
        filename
            下一步中会被分片上传的资源的完整名（包括扩展名）
        is_zipfile
            是否为zip文件
        file_path
            资源文件路径
        size
            资源文件大小
        creator_id
            创建者ID
        node_id
            节点ID
        owner
            资源的所有者，以逗号分隔的这些用户的ID拼成的字符串，无多余空格
        load_detail
            初始化时是否加载详情
        """
        self.name = name
        self.description = description
        self.state = state
        self.visibility = visibility
        self.upload_time = upload_time
        self.filename = filename
        self.is_zipfile = is_zipfile
        self.file_path = file_path
        self.size = size
        self.creator_id = creator_id
        self.node_id = node_id
        self.owner = owner
        super().__init__(id=id, load_detail=load_detail)

    @classmethod
    def list_dir(cls, resource_id):
        """
        文件列表查询接口

        :param resource_id: :obj:`str`
                    文件ID

        :return: 
            .. code-block:: json

                {
                  "vgg_ssd300_tianyuan.yaml": {
                        "name": "vgg_ssd300_tianyuan.yaml",
                        "type": "file"
                    }
                }

        """
        assert resource_id
        return get_with_token(f"{url_base()}/resource/listdir",
                              params={ 'file_id': resource_id })
    
    @classmethod
    def upload_file(cls,
                    resource_id: str,
                    file_path: str,
                    uploader: Optional[ResourceUploader]=None,
                    chunk_size: int=2048000):
        """
        对指定路径的本地文件进行分割并使用aiohttp异步上传

        Parameters
        ----------
        resource_id : :obj:`str`
            资源ID。
        file_path : :obj:`str`
            文件路径。
        uploader : :obj:`ResourceUploader`
            可以使用SDK中的AsyncResourceUploader，也可以自定义实现ResourceUploader。
        chunk_size : :obj:`int`
            文件分割大小，默认2048000。

        Returns
        -------
        bool
            True or False
        """

        assert resource_id, file_path

        # 读取选定文件的内容并根据chunk_size进行切割
        file_size = os.path.getsize(file_path)
        n_chunks = (file_size // chunk_size) + 1
        with open(file_path, 'rb') as f:
            chunks = [f.read(chunk_size) for i in range(n_chunks)]
        
        # 执行异步上传
        uploader.run(resource_id=resource_id, chunks=chunks)

        # 告知后端上传结束
        res = post_with_token(f"{url_base()}/resource/upload_finish",
                              data={ 'file_id': resource_id })
        return not not res

    @classmethod
    def download_file(cls,
                      resource_id: str,
                      save_path: str,
                      polling: Union[float, int]=5,
                      downloader: Optional[ResourceDownloader]=None,
                      ):
        """
        把服务器资源使用aiohttp异步下载到本地指定的文件夹

        Parameters
        ----------
        resource_id : :obj:`str`
            资源ID。
        save_path : :obj:`str`
            保存路径。
        downloader : :obj:`ResourceDownloader`
            可以使用SDK中的AsyncResourceDownloader，也可以自定义实现ResourceDownloader。
        polling : :obj:`float, int`
            下载前要先压缩文件，轮询查看文件有没有压缩完的时间间隔，单位：秒。默认值5

        Returns
        -------
        str
            文件名。
        """

        assert resource_id, save_path
        if not os.path.exists(save_path):
            raise AnyLearnException(f"保存路径{save_path}不存在")

        # 执行异步下载
        return downloader.run(resource_id=resource_id,  # type: ignore
                              save_path=save_path,
                              polling=polling,
                             )


class ResourceState:
    """
    资源状态标识：

    - 1(CREATED)表示已创建
    - 2(UPLOADING)表示上传中
    - 3(READY)表示就绪
    - -1(DELETED)表示已删除
    - -2(ERROR)表示出错
    """
    CREATED = 1
    UPLOADING = 2
    READY = 3
    DELETED = -1
    ERROR = -2


class ResourceVisibility:
    """
    资源可见性标识：

    - -1(HIDDEN)表示不可见
    - 1(PRIVATE)表示仅创建者可见
    - 2(PROTECTED)表示所有者可见
    - 3(PUBLIC)表示公开
    """
    HIDDEN = -1
    PRIVATE = 1
    PROTECTED = 2
    PUBLIC = 3



import os
from unittest.mock import patch, create_autospec
import sys

import responses

from anylearn.interfaces.resource.resource import Resource
from anylearn.utils.errors import AnyLearnException
from tests.base_test_case import BaseTestCase

class TestResource(BaseTestCase):
    @responses.activate
    def test_resource_list_dir(self):
        responses.add(responses.GET, url=self._url("resource/listdir?file_id=FILE001"),
                      match_querystring=True,
                      json={
                          'dir0': {
                              'name': "dir0",
                              'type': "dir",
                              'child': {
                                  'file0': {
                                      'name': "file0",
                                      'type': "file",
                                  },
                                  'file1': {
                                      'name': "file1",
                                      'type': "file",
                                  },
                              }
                          }
                      },
                      status=200)
        res = Resource.list_dir("FILE001")
        self.assertIn('dir0', res)
        self.assertIn('name', res['dir0'])
        self.assertIn('type', res['dir0'])
        self.assertIn('child', res['dir0'])
        self.assertEqual(res['dir0']['name'], "dir0")
        self.assertEqual(res['dir0']['type'], "dir")
        self.assertEqual(len(res['dir0']['child']), 2)
        self.assertIn('file0', res['dir0']['child'])
        self.assertIn('file1', res['dir0']['child'])
        self.assertEqual(res['dir0']['child']['file0']['name'], "file0")
        self.assertEqual(res['dir0']['child']['file1']['name'], "file1")
        self.assertEqual(res['dir0']['child']['file0']['type'], "file")
        self.assertEqual(res['dir0']['child']['file1']['type'], "file")

    @patch('anylearn.interfaces.resource.resource_uploader.ResourceUploader')
    @responses.activate
    def test_resource_upload_file(self, MockResourceUploader):
        @create_autospec
        def mock_upload_ok(resource_id, chunks):
            return True

        # Mock uploader with stubbed upload function
        uploader = MockResourceUploader()
        uploader.run = mock_upload_ok

        # Stub finish upload request
        responses.add(responses.POST, url=self._url("resource/upload_finish"),
                     json={
                         'msg': "上传已完成，稍后可查看上传结果"
                     },
                     status=200)
        
        # Stub a test file
        filename = "tests/testfile.txt"
        with open(filename, 'w') as f:
            f.write("test")
        
        res = Resource.upload_file("FILE001", filename, uploader=uploader)

        self.assertTrue(res)
        uploader.run.assert_called_once()

        os.remove(filename)

    @patch('anylearn.interfaces.resource.resource_downloader.ResourceDownloader')
    @responses.activate
    def test_resource_download_file(self, MockResourceDownloader):
        @create_autospec
        def mock_download_ok(resource_id, polling, save_path):
            return True
        
        downloader = MockResourceDownloader()
        downloader.run = mock_download_ok

        res = Resource.download_file(resource_id="FILE001",
                                     save_path="tests/",
                                     downloader=downloader,
                                    )

        self.assertTrue(res)
        downloader.run.assert_called_once()
        
    @patch('anylearn.interfaces.resource.resource_downloader.ResourceDownloader')
    @responses.activate
    def test_resource_download_file_savepath_not_exists(self, MockResourceDownloader):
        @create_autospec
        def mock_download_ok(resource_id,save_path):
            return True
        
        downloader = MockResourceDownloader()
        downloader.run = mock_download_ok

        save_path = "/data/wtt/aaa888"
        with self.assertRaises(AnyLearnException) as ctx:
            res = Resource.download_file("FILE001", save_path, downloader=downloader)
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, f"保存路径{save_path}不存在")



import click

from anylearn.cli.anylearn_cli_config import AnylearnCliConfig
from anylearn.cli.utils import (
    check_config,
    check_connection,
    cmd_error,
    cmd_success,
    get_cmd_command,
)
from anylearn.interfaces import Dataset


@click.group("dataset")
@check_config()
def commands():
    """
    Add local or remote dataset to local Anylearn project.
    """
    pass


@commands.command()
@get_cmd_command()
@click.argument('name')
@click.option(
    '-d', '--dir',
    prompt=True,
    help="Local dataset folder (absolute path)."
)
def local(name: str, dir: str):
    """
    Add local algorithm to current project.
    """
    dset = Dataset(name=name)
    config = AnylearnCliConfig.load()
    old_dset = config.datasets.get(name, None)
    old_dir = config.path['dataset'].get(name, None)
    if old_dset:
        cmd_error(msg=(
            "A dataset with same name "
            "has already been added to current project"
            "\n(\n"
            f"  ID={old_dset.id},\n"
            f"  NAME={old_dset.name},\n"
            f"  LOCAL_DIR={old_dir},\n"
            "\n)\n"
        ))
        raise click.Abort
    config.datasets[name] = dset
    config.path['dataset'][name] = dir
    AnylearnCliConfig.update(config)
    cmd_success(msg="ADDED")

@commands.command()
@click.argument('id')
@check_connection()
@get_cmd_command()
def remote(id: str):
    """
    Add remote dataset by ID to current project.
    """
    config = AnylearnCliConfig.load()
    try:
        old_dset = next(
            d
            for d in config.datasets.values()
            if d.id == id
        )
        cmd_error(msg=(
            f"Remote dataset (ID={id}, name={old_dset.name}) "
            "has already been added to current project.\n"
            "Aborted!"
        ))
    except:
        try:
            dset = Dataset(id=id, load_detail=True)
            config = AnylearnCliConfig.load()
            old_dset = config.datasets.get(dset.name, None) # type: ignore
            old_dir = config.path['dataset'].get(dset.name, None) # type: ignore
            if old_dset:
                cmd_error(msg=(
                    "A dataset with same name "
                    "has already been added to current project"
                    "\n(\n"
                    f"  ID={old_dset.id},\n"
                    f"  NAME={old_dset.name},\n"
                    f"  LOCAL_DIR={old_dir},\n"
                    "\n)\n"
                ))
                raise click.Abort
            config.datasets[dset.name] = dset # type: ignore
            config.path['dataset'][dset.name] = None # type: ignore
            AnylearnCliConfig.update(config)
            cmd_success(msg="ADDED")
        except:
            cmd_error(msg=(
                f"Remote dataset (ID={id}) does not exist.\n"
                "Aborted!"
            ))



from anylearn.interfaces.train_task import TrainTask
import os
from typing import Optional

import click

import anylearn.cli.add
import anylearn.cli.config
import anylearn.cli.download
import anylearn.cli.push
import anylearn.cli.quota
import anylearn.cli.rm
import anylearn.cli.run
from anylearn.cli.run import (
    _check_algorithm_name,
    _check_project,
    _do_train,
    _load_training_config,
)
import anylearn.cli.task
import anylearn.cli.view
from anylearn.cli.anylearn_cli_config import AnylearnCliConfig
from anylearn.cli.utils import (
    check_config,
    get_history_file_path,
    get_cmd_command,
    cmd_error,
    cmd_info,
    cmd_success,
    cmd_warning,
    check_connection,
    do_login,
    do_logout,
)
from anylearn.interfaces import Project


@click.group()
@click.version_option()
def cli():
    pass


@cli.command()
@get_cmd_command()
@click.option(
    '-n', '--project-name',
    default=os.path.basename(os.getcwd()),
    help="Project name.",
)
@click.option(
    '-d', '--project-description',
    default=None,
    help="Project description.",
)
@click.option(
    '--login',
    is_flag=True,
    default=False,
    help="Prompt login in the meantime.",
)
def init(
    project_name: str=os.path.basename(os.getcwd()),
    project_description: str=None,
    login: bool=False,
):
    """
    Create an empty Anylearn project.
    """
    if login:
        do_login(show_success=True)
    config = AnylearnCliConfig.load()
    if config.project and config.project.name:
        cmd_warning(msg=(
            "Project ("
            f"ID={config.project.id}, "
            f"name={config.project.name}, "
            f"description={config.project.description}, "
            ") "
            "has already been initialized in current directory.\n"
            "Call `anyctl push project` to synchronize local project "
            "or `anyctl reset` to reset.\n"
            "Aborted!"
        ))
        return
    config.project = Project(
        name=project_name,
        description=project_description,
    )
    AnylearnCliConfig.update(config)
    cmd_success(msg="Initialized Anylearn project:")
    cmd_info(msg=str(config))


@cli.command()
@check_config()
@get_cmd_command()
def reset():
    """
    Reset current Anylearn project.
    """
    cmd_warning(msg="Project in current directory will be reverted.")
    if click.confirm("Are you sure you want to proceed?"):
        AnylearnCliConfig.get_config_path().unlink()
        cmd_success(msg=(
            "Local project has been reverted.\n"
            "However, runtime files such as <hyperparams>.json are kept. "
            "You can still make use of them in other projects, "
            "or remove them manually when no longer needed."
        ))


@cli.command()
@check_config()
@get_cmd_command()
def login():
    do_login(show_success=True)


@cli.command()
@check_config()
@get_cmd_command()
def logout():
    do_logout()


@cli.command()
@click.option(
    '-f', '--file',
    default="AnylearnTraining.yaml",
    help="""
    Path of training configuration file 
    (default is './AnylearnTraining.yaml')
    .
    """
)
@check_config()
@check_connection()
def train(file: str="AnylearnTraining.yaml"):
    """
    Run training task based on configuration file.
    """
    config = AnylearnCliConfig.load()
    _check_project(config)
    training_config = _load_training_config(file)
    _check_algorithm_name(config, training_config['algorithm_name'])
    task = _do_train(config=config, **training_config)
    cmd_success(msg=f"CREATED {task.id}")


@cli.command()
@click.option(
    '-i', '--task-id',
    type=str,
    help="""
    ID of a certain task to fetch logs.
    If not specified, the last training task's ID will be used.
    """
)
@click.option(
    '-l', '--limit',
    type=int,
    default=100,
    show_default=True,
    help="""
    Max number of lines of logs to fetch.
    """
)
@click.option(
    '-f', '--follow',
    is_flag=True,
    default=False,
    help="""
    Whether the logs should be streamed and printed as they grow.
    """
)
@click.option(
    '--full-export',
    is_flag=True,
    default=False,
    help="""
    Fetch and save full logs into file.
    """
)
@check_config()
@check_connection()
def log(task_id: Optional[str]=None,
        limit: int=100,
        follow: bool=False,
        full_export: bool=False):
    """
    Show or export logs of training task.
    """
    config = AnylearnCliConfig.load()
    _check_project(config)
    if not task_id:
        tasks = config.project.get_train_tasks()
        if not tasks:
            cmd_error(msg=f"No launched training tasks in current project.")
            raise click.Abort
        task_id = sorted(tasks, key=lambda t: t.create_time)[0].id
    if not task_id.startswith('TRAI'):
        cmd_error(msg=f"Command `log` supports training tasks only.")
        raise click.Abort
    task = TrainTask(id=task_id)
    if full_export:
        logs = "\r\n".join(task.get_full_log())
        filename = f"{task_id}.log"
        with open(filename, 'w') as f:
            f.write(logs)
        cmd_success(msg=(
            f"Full log of task {task_id} "
            f"is exported to file {filename}."
        ))
    elif follow:
        for line in task.stream_log(init_limit=limit):
            cmd_info(msg=line)
    else:
        cmd_info(msg="\r\n".join(task.get_last_log(limit=limit)))


@cli.command()
@click.option(
    '-t', '--type',
    type=click.Choice(["get", "clean"]),
    default="get",
    help="""
    Actions on history, default "get".
    """
)
def history(type: str):
    """
    Show project's task history.
    """
    if type == "clean":
        if click.confirm("All cmd history will be cleared, continue?"):
            get_history_file_path().unlink()
            cmd_info(msg="History has been cleared")
        else:
            cmd_info(msg="Cancel clean")
    else:
        with open(get_history_file_path(), 'r') as f:
            command = f.read()
        cmd_info(msg=command)


cli.add_command(anylearn.cli.add.commands)
cli.add_command(anylearn.cli.config.commands)
cli.add_command(anylearn.cli.download.commands)
cli.add_command(anylearn.cli.push.commands)
cli.add_command(anylearn.cli.quota.commands)
cli.add_command(anylearn.cli.rm.commands)
cli.add_command(anylearn.cli.run.commands)
cli.add_command(anylearn.cli.task.commands)
cli.add_command(anylearn.cli.view.commands)


if __name__ == '__main__':
    cli()



import json
from requests.exceptions import RequestException
import responses
from urllib.parse import urlencode

from anylearn.interfaces.evaluate_task import (
    EvaluateSubTask,
    EvaluateTask,
    EvaluateTaskState,
    EvaluateTaskVisibility
)
from anylearn.utils.errors import (
    AnyLearnException,
    AnyLearnMissingParamException
)
from tests.base_test_case import BaseTestCase

class TestEvaluateTask(BaseTestCase):
    @responses.activate
    def test_list_evaluate_task_ok(self):
        responses.add(responses.GET, url=self._url("evaluate_task/list"),
                      json=[
                          {
                              'id': "EVAL001",
                              'name': "TestEvalTask1",
                              'description': "test",
                              'state': EvaluateTaskState.SUCCESS,
                              'visibility': EvaluateTaskVisibility.PUBLIC,
                              'creator_id': "USER001",
                              'owner': ["USER001"],
                              'create_time': "2020-04-01 00:00",
                              'finish_time': "2020-04-02 00:00",
                              'secret_key': "SECRET",
                          },
                          {
                              'id': "EVAL002",
                              'name': "TestEvalTask2",
                              'description': "test",
                              'state': EvaluateTaskState.RUNNING,
                              'visibility': EvaluateTaskVisibility.PRIVATE,
                              'creator_id': "USER002",
                              'owner': ["USER002"],
                              'create_time': "2020-04-03 00:00",
                              'finish_time': "",
                              'secret_key': "SECRET",
                          },
                      ],
                      status=200)
        evals = EvaluateTask.get_list()
        self.assertIsInstance(evals, list)
        self.assertEqual(len(evals), 2)
        self.assertIsInstance(evals[0], EvaluateTask)
        self.assertIsInstance(evals[1], EvaluateTask)
        self.assertEqual(evals[0].id, "EVAL001")
        self.assertEqual(evals[1].id, "EVAL002")

    @responses.activate
    def test_list_evaluate_task_ko_unknown_response(self):
        responses.add(responses.GET, url=self._url("evaluate_task/list"),
                      json={'msg': "Unknown response"},
                      status=200)
        with self.assertRaises(AnyLearnException) as ctx:
            EvaluateTask.get_list()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_get_evaluate_task_detail_explicit_call_ok(self):
        responses.add(responses.GET, url=self._url("evaluate_task/query?id=EVAL001"),
                      match_querystring=True, json={
                          'id': "EVAL001",
                          'name': "TestEvaluateTask1",
                          'description': "test",
                          'state': EvaluateTaskState.RUNNING,
                          'visibility': 3,
                          'creator_id': "USER001",
                          'owner': ["USER001"],
                          'create_time': "2021-04-01 23:59:59",
                          'finish_time': "",
                          'params_list': '[{"model_id": "MODE001", "files": "DSET001", "evaluate_params": {"dataset": "$DSET001", "model_path": "$MODE001"}}]',
                          'secret_key': "SECRET",
                          'results': [
                              {
                                  'args': '{"dataset": "$DSET001", "model_path": "$MODE001"}',
                                  'envs': "",
                                  'files': "DSET001",
                                  'gpu_num': 2,
                                  'id': "EVRE001",
                                  'model_id': "MODE001",
                                  'results': "",
                              }
                          ],
                      },
                      status=200)
        eval_task = EvaluateTask(id="EVAL001")
        eval_task.get_detail()
        self.assertEqual(eval_task.name, "TestEvaluateTask1")
        self.assertEqual(eval_task.description, "test")
        self.assertEqual(eval_task.state, EvaluateTaskState.RUNNING)
        self.assertEqual(eval_task.visibility, EvaluateTaskVisibility.PUBLIC)
        self.assertEqual(eval_task.creator_id, "USER001")
        self.assertIsInstance(eval_task.owner, list)
        self.assertEqual(eval_task.create_time, "2021-04-01 23:59:59")
        self.assertEqual(eval_task.secret_key, "SECRET")
        self.assertIsInstance(eval_task.sub_tasks, list)
        if isinstance(eval_task.sub_tasks, list):
            self.assertEqual(len(eval_task.sub_tasks), 1)
        self.assertEqual(eval_task.sub_tasks[0].id, "EVRE001")
        self.assertEqual(eval_task.sub_tasks[0].model_id, "MODE001")
        self.assertEqual(eval_task.sub_tasks[0].evaluate_params['dataset'], "$DSET001")
        self.assertEqual(eval_task.sub_tasks[0].evaluate_params['model_path'], "$MODE001")
        self.assertEqual(eval_task.sub_tasks[0].files[0], "DSET001")
        self.assertFalse(eval_task.sub_tasks[0].envs)
        self.assertEqual(eval_task.sub_tasks[0].gpu_num, 2)
        self.assertFalse(eval_task.sub_tasks[0].results)

    @responses.activate
    def test_get_evaluate_task_detail_no_explicit_call_ok(self):
        responses.add(responses.GET, url=self._url("evaluate_task/query?id=EVAL002"),
                      match_querystring=True, json={
                          'id': "EVAL002",
                          'name': "TestEvaluateTask2",
                          'description': "test",
                          'state': EvaluateTaskState.RUNNING,
                          'visibility': 3,
                          'creator_id': "USER002",
                          'owner': ["USER002"],
                          'create_time': "2021-04-02 23:59:59",
                          'finish_time': "",
                          'params_list': '[{"model_id": "MODE002", "files": "DSET002", "evaluate_params": {"dataset": "$DSET002", "model_path": "$MODE002"}}]',
                          'secret_key': "SECRET",
                          'results': [
                              {
                                  'args': '{"dataset": "$DSET002", "model_path": "$MODE002"}',
                                  'envs': "",
                                  'files': "DSET002",
                                  'gpu_num': 1,
                                  'id': "EVRE002",
                                  'model_id': "MODE002",
                                  'results': "",
                              }
                          ],
                      },
                      status=200)
        eval_task = EvaluateTask(id="EVAL002", load_detail=True)
        self.assertEqual(eval_task.name, "TestEvaluateTask2")
        self.assertEqual(eval_task.description, "test")
        self.assertEqual(eval_task.state, EvaluateTaskState.RUNNING)
        self.assertEqual(eval_task.visibility, EvaluateTaskVisibility.PUBLIC)
        self.assertEqual(eval_task.creator_id, "USER002")
        self.assertIsInstance(eval_task.owner, list)
        self.assertEqual(eval_task.create_time, "2021-04-02 23:59:59")
        self.assertEqual(eval_task.secret_key, "SECRET")
        self.assertIsInstance(eval_task.sub_tasks, list)
        if isinstance(eval_task.sub_tasks, list):
            self.assertEqual(len(eval_task.sub_tasks), 1)
        self.assertEqual(eval_task.sub_tasks[0].id, "EVRE002")
        self.assertEqual(eval_task.sub_tasks[0].model_id, "MODE002")
        self.assertEqual(eval_task.sub_tasks[0].evaluate_params['dataset'], "$DSET002")
        self.assertEqual(eval_task.sub_tasks[0].evaluate_params['model_path'], "$MODE002")
        self.assertEqual(eval_task.sub_tasks[0].files[0], "DSET002")
        self.assertFalse(eval_task.sub_tasks[0].envs)
        self.assertEqual(eval_task.sub_tasks[0].gpu_num, 1)
        self.assertFalse(eval_task.sub_tasks[0].results)

    @responses.activate
    def test_get_evaluate_task_detail_ko_403(self):
        responses.add(responses.GET, url=self._url("evaluate_task/query?id=EVAL403"),
                      match_querystring=True, status=403)
        evaluate_task = EvaluateTask(id="EVAL403")
        with self.assertRaises(RequestException) as ctx:
            evaluate_task.get_detail()
        e = ctx.exception
        self.assertIsInstance(e, RequestException)
        self.assertEqual(e.response.status_code, 403)

    @responses.activate
    def test_get_evaluate_task_detail_ko_unknown_response(self):
        responses.add(responses.GET, url=self._url("evaluate_task/query?id=EVAL250"),
                      match_querystring=True, json=[{'msg': "Unknown response"}],
                      status=200)
        evaluate_task = EvaluateTask(id="EVAL250")
        with self.assertRaises(AnyLearnException) as ctx:
            evaluate_task.get_detail()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_create_evaluate_task_ok(self):
        responses.add(responses.POST, url=self._url("evaluate_task/add"),
                      json={'data': "EVAL001", 'message': "服务添加成功"},
                      status=200)
        evaluate_task = EvaluateTask(name="TestEvalTask001",
                                     sub_tasks=[
                                         EvaluateSubTask(
                                             model_id="MODE001",
                                             evaluate_params={
                                                 'dataset': "$DSET001",
                                                 'pre_train': "$MODE002",
                                                 'model_path': "$MODE001",
                                             },
                                             files=["DSET001", "MODE002"])
                                     ])
        result = evaluate_task.save()
        self.assertTrue(result)
        self.assertEqual(evaluate_task.id, "EVAL001")
        self.assertEqual(evaluate_task.params_list, '[{"id": null, "model_id": "MODE001", "files": "DSET001,MODE002", "evaluate_params": {"dataset": "$DSET001", "pre_train": "$MODE002", "model_path": "$MODE001"}, "envs": "", "gpu_num": 1, "results": null}]')

    def test_create_evaluate_task_ko_empty_name_subtasks(self):
        evaluate_task = EvaluateTask()
        with self.assertRaises(AnyLearnMissingParamException) as ctx:
            evaluate_task.save()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnMissingParamException)
        self.assertEqual(e.msg, "EvaluateTask缺少必要字段：['name', 'sub_tasks']")

    def test_create_evaluate_task_ko_empty_subtaskmodelid(self):
        evaluate_task = EvaluateTask(name="TestEvalTask",
                                     sub_tasks=[
                                         EvaluateSubTask(
                                             evaluate_params={
                                                 'dataset': "$DSET001",
                                                 'pre_train': "$MODE002",
                                             },
                                             files=["DSET001", "MODE002"])
                                     ])
        with self.assertRaises(AnyLearnMissingParamException) as ctx:
            evaluate_task.save()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnMissingParamException)
        self.assertEqual(e.msg, "EvaluateSubTask缺少必要字段：model_id")

    @responses.activate
    def test_create_evaluate_task_ko_unknown_response(self):
        responses.add(responses.POST, url=self._url("evaluate_task/add"),
                      json={'msg': "Unknown response"}, status=201)
        evaluate_task = EvaluateTask(name="TestEvalTask250",
                                     sub_tasks=[
                                         EvaluateSubTask(
                                             model_id="MODE250",
                                             evaluate_params={
                                                 'dataset': "$DSET001",
                                             },
                                             files=["DSET001"])
                                     ])
        with self.assertRaises(AnyLearnException) as ctx:
            evaluate_task.save()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_delete_evaluate_task_ok(self):
        responses.add(responses.DELETE,
                      url=self._url("evaluate_task/delete?id=EVAL001&force=0"),
                      match_querystring=True,
                      json={'data': "EVAL001", 'message': "任务删除成功"},
                      status=200)
        evaluate_task = EvaluateTask(id="EVAL001")
        result = evaluate_task.delete()
        self.assertTrue(result)

    @responses.activate
    def test_delete_evaluate_task_ko_unknown_response(self):
        responses.add(responses.DELETE,
                      url=self._url("evaluate_task/delete?id=EVAL250&force=0"),
                      match_querystring=True,
                      json={'msg': "Unknown response"},
                      status=204)
        evaluate_task = EvaluateTask(id="EVAL250")
        with self.assertRaises(AnyLearnException) as ctx:
            evaluate_task.delete()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_get_evaluate_task_log_ok(self):
        query_str = urlencode({
            'id': "EVAL001",
            'limit': 100,
            'direction': "init",
            'index': 0,
            'self': 1,
        })
        responses.add(responses.GET,
                      url=self._url(f"evaluate_task/log?{query_str}"),
                      match_querystring=True,
                      json={
                          'EVRE001': [
                            {'offset': 1, 'text': "log1"},
                            {'offset': 2, 'text': "log2"},
                            {'offset': 3, 'text': "log3"},
                          ],
                          'EVRE002': [
                            {'offset': 4, 'text': "log4"},
                            {'offset': 5, 'text': "log5"},
                          ],
                      },
                      status=200)
        evaluate_task = EvaluateTask(id="EVAL001")
        result = evaluate_task.get_log(includes=True)
        self.assertIsInstance(result, dict)
        self.assertEqual(len(result), 2)
        self.assertIn('EVRE001', result)
        self.assertIn('EVRE002', result)
        self.assertIsInstance(result['EVRE001'], list)
        self.assertIsInstance(result['EVRE002'], list)
        self.assertEqual(len(result['EVRE001']), 3)
        self.assertEqual(len(result['EVRE002']), 2)
        self.assertIn('text', result['EVRE001'][0])
        self.assertIn('text', result['EVRE001'][1])
        self.assertIn('text', result['EVRE001'][2])
        self.assertIn('text', result['EVRE002'][0])
        self.assertIn('text', result['EVRE002'][1])

    def test_get_evaluate_task_log_ko_empty_id(self):
        evaluate_task = EvaluateTask(name="TestEvalTask")
        with self.assertRaises(AnyLearnMissingParamException) as ctx:
            evaluate_task.get_log()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnMissingParamException)
        self.assertEqual(e.msg, "EvaluateTask缺少必要字段：['id']")

    @responses.activate
    def test_get_evaluate_task_log_ko_unknown_response(self):
        query_str = urlencode({
            'id': "EVAL250",
            'limit': 100,
            'direction': "init",
            'index': 0,
            'self': 0,
        })
        responses.add(responses.GET,
                      url=self._url(f"evaluate_task/log?{query_str}"),
                      match_querystring=True,
                      json="",
                      status=204)
        evaluate_task = EvaluateTask(id="EVAL250")
        with self.assertRaises(AnyLearnException) as ctx:
            evaluate_task.get_log()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")



import click

from anylearn.cli.anylearn_cli_config import AnylearnCliConfig
from anylearn.cli.utils import (
    check_config,
    check_connection,
    cmd_error,
    cmd_success,
    get_cmd_command,
)
from anylearn.interfaces import Algorithm


@click.group("algorithm")
@check_config()
def commands():
    """
    Add local or remote algorithm to local Anylearn project.
    """
    pass


@commands.command()
@get_cmd_command()
@click.argument('name')
@click.option(
    '-d', '--dir',
    prompt=True,
    help="Local algorithm folder (absolute path)."
)
@click.option(
    "--entrypoint-training",
    prompt=True,
    help="Training entrypoint command of algorithm."
)
# @click.option(
#     "--entrypoint-evaluation",
#     prompt=True,
#     help="Evaluation entrypoint command of algorithm."
# )
@click.option(
    "--output-training",
    prompt=True,
    help="Training model saving path of algorithm."
)
# @click.option(
#     "--output-evaluation",
#     prompt=True,
#     help="Evaluation results saving path of algorithm."
# )
@click.option(
    '-i', '--image',
    default='QUICKSTART',
    help="Container image name.",
)
def local(
    name: str,
    dir: str,
    entrypoint_training: str,
    # entrypoint_evaluation: str,
    output_training: str,
    # output_evaluation: str,
    image: str='QUICKSTART',
):
    """
    Add local algorithm to current project.
    """
    algo = Algorithm(
            name=name,
            train_params = "[]",
            evaluate_params = "[]",
            follows_anylearn_norm=False,
            entrypoint_training=entrypoint_training,
            output_training=output_training,
            # entrypoint_evaluation=entrypoint_evaluation,
            # output_evaluation=output_evaluation,
    )
    config = AnylearnCliConfig.load()
    old_algo = config.algorithms.get(name, None)
    old_dir = config.path['algorithm'].get(name, None)
    if old_algo:
        cmd_error(msg=(
            "An algorithm with same name "
            "has already been added to current project"
            "\n(\n"
            f"  ID={old_algo.id},\n"
            f"  NAME={old_algo.name},\n"
            f"  LOCAL_DIR={old_dir},\n"
            "\n)\n"
        ))
        raise click.Abort
    config.algorithms[name] = algo
    config.path['algorithm'][name] = dir
    config.images[name] = image
    AnylearnCliConfig.update(config)
    cmd_success(msg="ADDED")


@commands.command()
@click.argument('id')
@check_connection()
@get_cmd_command()
def remote(id: str):
    """
    Add remote algorithm by ID to current project.
    """
    config = AnylearnCliConfig.load()
    try:
        old_algo = next(
            a
            for a in config.algorithms.values()
            if a.id == id
        )
        cmd_error(msg=(
            f"Remote algorithm (ID={id}, name={old_algo.name}) "
            "has already been added to current project.\n"
            "Aborted!"
        ))
    except:
        try:
            algo = Algorithm(id=id, load_detail=True)
            config = AnylearnCliConfig.load()
            old_algo = config.algorithms.get(algo.name, None) # type: ignore
            old_dir = config.path['algorithm'].get(algo.name, None) # type: ignore
            if old_algo:
                cmd_error(msg=(
                    "An algorithm with same name "
                    "has already been added to current project"
                    "\n(\n"
                    f"  ID={old_algo.id},\n"
                    f"  NAME={old_algo.name},\n"
                    f"  LOCAL_DIR={old_dir},\n"
                    "\n)\n"
                ))
                raise click.Abort
            config.algorithms[algo.name] = algo # type: ignore
            config.path['algorithm'][algo.name] = None # type: ignore
            AnylearnCliConfig.update(config)
            cmd_success(msg="ADDED")
        except:
            cmd_error(msg=(
                f"Remote algorithm (ID={id}) does not exist.\n"
                "Aborted!"
            ))



from __future__ import annotations
from base64 import b64decode, b64encode
from io import StringIO
import json
import os
from pathlib import Path
from typing import Optional

from ruamel.yaml import YAML

from anylearn.cli.errors import ConfigurationKeyError, NameConflictError
from anylearn.interfaces import (
    Algorithm,
    Dataset,
    Project,
    ResourceVisibility,
)


class AnylearnCliConfig:

    remote: dict = {
        'host': None,
        'username': None,
        'password': None,
    }
    project: Optional[Project] = None
    algorithms: dict[str, Algorithm] = {}
    datasets: dict[str, Dataset] = {}
    path: dict[str, dict[str, str]] = {
        'algorithm': {},
        'dataset': {},
    }
    images: dict[str, str] = {}

    def gets(self, config_key: str):
        funcs, keys = self.__parse_config_key(config_key=config_key)
        try:
            return funcs['get'](keys=keys)
        except Exception as e:
            if not str(e):
                raise ConfigurationKeyError(f"Unknown configuration key {config_key}")
            raise

    def sets(self, config_key: str, value):
        funcs, keys = self.__parse_config_key(config_key=config_key)
        try:
            funcs['set'](keys=keys, value=value)
        except Exception as e:
            if not str(e):
                raise ConfigurationKeyError(f"Unknown configuration key {config_key}")
            raise

    def __parse_config_key(self, config_key: str):
        parts = config_key.split('.')
        if not parts:
            raise ConfigurationKeyError("Configuration key required.")
        meta_key = parts[0]
        mapping = {
            'project': {
                'get': self.__get_in_project,
                'set': self.__set_in_project,
            },
            'algorithms': {
                'get': self.__get_in_algorithms,
                'set': self.__set_in_algorithms,
            },
            'datasets': {
                'get': self.__get_in_datasets,
                'set': self.__set_in_datasets,
            },
            'path': {
                'get': self.__get_in_path,
                'set': self.__set_in_path,
            },
            'images': {
                'get': self.__get_in_images,
                'set': self.__set_in_images,
            }
        }
        if meta_key not in mapping.keys():
            raise ConfigurationKeyError(f"Unknown configuration key {config_key}")
        del(parts[0])
        return mapping[meta_key], parts

    def __get_in_project(self, keys: list):
        if not keys or len(keys) != 1:
            raise ConfigurationKeyError
        return getattr(self.project, keys[0])

    def __set_in_project(self, keys: list, value):
        if not keys or len(keys) != 1:
            raise ConfigurationKeyError
        setattr(self.project, keys[0], value)

    def __get_in_algorithms(self, keys: list):
        if not keys or len(keys) != 2:
            raise ConfigurationKeyError
        return getattr(self.algorithms[keys[0]], keys[1])

    def __set_in_algorithms(self, keys: list, value):
        if not keys or len(keys) != 2:
            raise ConfigurationKeyError
        if keys[1] == 'name':
            if value in self.algorithms:
                raise NameConflictError(
                    f"Failed to rename algorithm {keys[0]} to {value}: "
                    f"algorithm {value} already added in current project."
                )
            self.algorithms[keys[0]].name = value
            self.algorithms[value] = self.algorithms[keys[0]]
            self.path['algorithm'][value] = self.path['algorithm'][keys[0]]
            del(self.algorithms[keys[0]])
            del(self.path['algorithm'][keys[0]])
        else:
            setattr(self.algorithms[keys[0]], keys[1], value)

    def __get_in_datasets(self, keys: list):
        if not keys or len(keys) != 2:
            raise ConfigurationKeyError
        return getattr(self.datasets[keys[0]], keys[1])

    def __set_in_datasets(self, keys: list, value):
        if not keys or len(keys) != 2:
            raise ConfigurationKeyError
        if keys[1] == 'name':
            if value in self.datasets:
                raise NameConflictError(
                    f"Failed to rename dataset {keys[0]} to {value}: "
                    f"dataset {value} already added in current project."
                )
            self.datasets[keys[0]].name = value
            self.datasets[value] = self.datasets[keys[0]]
            self.path['dataset'][value] = self.path['dataset'][keys[0]]
            del(self.datasets[keys[0]])
            del(self.path['dataset'][keys[0]])
        else:
            setattr(self.datasets[keys[0]], keys[1], value)

    def __get_in_path(self, keys: list):
        if not keys or len(keys) != 2 or not keys[0] in self.path:
            raise ConfigurationKeyError
        return self.path[keys[0]][keys[1]]

    def __set_in_path(self, keys: list, value: str):
        if not keys or len(keys) != 2 or not keys[0] in self.path:
            raise ConfigurationKeyError
        self.path[keys[0]][keys[1]] = value

    def __get_in_images(self, keys: list):
        if not keys or len(keys) != 1 or not keys[0] in self.images:
            raise ConfigurationKeyError
        return self.images[keys[0]]

    def __set_in_images(self, keys: list, value: str):
        if not keys or len(keys) != 1 or not keys[0] in self.images:
            raise ConfigurationKeyError
        self.images[keys[0]] = value

    def __str__(self) -> str:
        yaml = YAML()
        with StringIO() as stream:
            yaml.dump(AnylearnCliConfig.serialize(config=self), stream)
            s = stream.getvalue()
        return s

    @staticmethod
    def serialize(config: AnylearnCliConfig) -> dict:
        config_dict = {}
        config_dict['auth'] = AnylearnCliConfig.remote2auth(config.remote)
        config_dict['project'] = {} if not config.project else {
            'id': config.project.id,
            'name': config.project.name,
            'description': config.project.description,
        }
        config_dict['algorithms'] = {
            algo_name: {
                'id': algo.id,
                'name': algo.name,
                'description': algo.description,
                'visibility': algo.visibility,
                'train_params': algo.train_params,
                'evaluate_params': algo.evaluate_params,
                'follows_anylearn_norm': algo.follows_anylearn_norm,
                'entrypoint_training': algo.entrypoint_training,
                'entrypoint_evaluation': algo.entrypoint_evaluation,
                'output_training': algo.output_training,
                'output_evaluation': algo.output_evaluation,
            }
            for algo_name, algo in config.algorithms.items()
        }
        config_dict['datasets'] = {
            dset_name: {
                'id': dset.id,
                'name': dset.name,
                'description': dset.description,
                'visibility': dset.visibility,
            }
            for dset_name, dset in config.datasets.items()
        }
        config_dict['path'] = config.path
        config_dict['images'] = config.images
        return config_dict

    @staticmethod
    def deserialize(config_dict: dict) -> AnylearnCliConfig:
        proj_dict = config_dict.get('project', {})
        algorithms = config_dict.get('algorithms', {})
        datasets = config_dict.get('datasets', {})
        path = config_dict.get('path', {
            'algorithm': {},
            'dataset': {},
        })
        images = config_dict.get('images', {})
        config = AnylearnCliConfig()
        config.remote = AnylearnCliConfig.auth2remote(config_dict['auth'])
        config.project = Project(
            id=proj_dict.get('id', None),
            name=proj_dict.get('name', None),
            description=proj_dict.get('description', None),
        )
        config.algorithms = {
            algo_name: Algorithm(
                id=algo_dict.get('id', None),
                name=algo_dict.get('name', None),
                description=algo_dict.get('description', None),
                visibility=algo_dict.get('visibility',
                                         ResourceVisibility.PRIVATE),
                train_params=json.dumps(algo_dict.get('train_params', [])),
                evaluate_params=json.dumps(algo_dict.get('evaluate_params', [])),
                follows_anylearn_norm=algo_dict.get('follows_anylearn_norm', True),
                entrypoint_training=algo_dict.get('entrypoint_training', None),
                entrypoint_evaluation=algo_dict.get('entrypoint_evaluation', None),
                output_training=algo_dict.get('output_training', None),
                output_evaluation=algo_dict.get('output_evaluation', None),
            )
            for algo_name, algo_dict in algorithms.items()
        }
        config.datasets = {
            dset_name: Dataset(
                id=dset_dict.get('id', None),
                name=dset_dict.get('name', None),
                description=dset_dict.get('description', None),
                visibility=dset_dict.get('visibility',
                                            ResourceVisibility.PRIVATE),
            )
            for dset_name, dset_dict in datasets.items()
        }
        config.path = path
        config.images = images
        return config

    @staticmethod
    def auth2remote(auth_str: str):
        if not auth_str:
            return {
                'host': None,
                'username': None,
                'password': None,
            }
        s = b64decode(auth_str.encode('utf8')).decode('utf8')
        cred, host = s.split('@')
        username, p = cred.split(':')
        password = b64decode(p.encode('utf8')).decode('utf8')
        return {
            'host': host,
            'username': username,
            'password': password,
        }

    @staticmethod
    def remote2auth(remote: dict):
        if not all([
            'host' in remote and remote['host'],
            'username' in remote and remote['username'],
            'password' in remote and remote['password'],
        ]):
            return ''
        p = b64encode(remote['password'].encode('utf8')).decode('utf8')
        s = f"{remote['username']}:{p}@{remote['host']}"
        return b64encode(s.encode('utf8')).decode('utf8')

    @staticmethod
    def get_config_path() -> Path:
        cwd = Path(os.getcwd())
        config_path = cwd / 'AnylearnProject.yaml'
        return config_path

    @staticmethod
    def load():
        config_path = AnylearnCliConfig.get_config_path()
        if not config_path.exists():
            return AnylearnCliConfig()
        yaml = YAML()
        return AnylearnCliConfig.deserialize(config_dict=yaml.load(config_path))

    @staticmethod
    def update(config: AnylearnCliConfig):
        config_path = AnylearnCliConfig.get_config_path()
        config_path.touch(exist_ok=True)
        yaml = YAML()
        yaml.dump(AnylearnCliConfig.serialize(config=config), config_path)



from multiprocessing import Queue
from pathlib import Path
import sys
from threading import Thread
import time
import traceback
from typing import Optional, Union

from .utils import (
    _get_archive_checksum,
    _get_or_create_resource_archive,
    make_name_by_path,
)
from ..interfaces.resource import (
    Dataset,
    Resource,
    ResourceState,
    ResourceUploader,
    ResourceVisibility,
    SyncResourceUploader,
)
from ..storage.db import DB
from ..utils import logger
from ..utils.errors import (
    AnyLearnException,
    AnyLearnMissingParamException,
)


def sync_dataset(
    id: Optional[str]=None,
    name: Optional[str]=None,
    dir_path: Optional[Union[str, Path]]=None,
    archive_path: Optional[str]=None,
    uploader: Optional[ResourceUploader]=None,
    polling: Union[float, int]=5,
) -> Dataset:
    try:
        return Dataset(id=id, load_detail=True)
    except:
        dset, archive, upload = _sync_local_dataset(
            name=name,
            dir_path=dir_path,
            archive_path=archive_path,
        )
        if archive and upload:
            _upload_dataset(
                dataset=dset,
                archive=archive,
                uploader=uploader,
                polling=polling,
            )
        return dset


def _sync_local_dataset(
    name: Optional[str]=None,
    dir_path: Optional[Union[str, Path]]=None,
    archive_path: Optional[str]=None,
):
    if not any([dir_path, archive_path]):
        raise AnyLearnMissingParamException((
            "None of ['dir_path', 'archive_path'] "
            "is specified."
        ))
    if not name:
        name = make_name_by_path(dir_path or archive_path)
    archive_path = _get_or_create_resource_archive(
        name=name,
        dir_path=dir_path,
        archive_path=archive_path
    )
    checksum = _get_archive_checksum(archive_path)
    dset, upload = _get_or_create_raw_dataset(name=name, checksum=checksum)
    return dset, archive_path, upload


def _get_or_create_raw_dataset(name, checksum):
    to_upload = False
    dset = _get_dataset_by_checksum(checksum)
    if not dset:
        to_upload = True
        dset = _create_new_dataset(name)
    return dset, to_upload


def _get_dataset_by_checksum(checksum) -> Optional[Dataset]:
    local_id = DB().find_local_dataset_by_checksum(checksum=checksum)
    if not local_id:
        return None
    try:
        return Dataset(id=local_id, load_detail=True)
    except:
        logger.warning(
            f"Local dataset ({local_id}) "
            "has been deleted remotely, "
            "forced to re-registering dataset."
        )
        DB().delete_local_dataset(id=local_id)
        return None


def _create_new_dataset(name) -> Dataset:
    dset = Dataset(
        name=name,
        description="SDK_QUICKSTART",
        visibility=ResourceVisibility.PRIVATE,
        filename=f"{name}.zip",
        is_zipfile=True,
    )
    dset.save()
    return dset


def _upload_dataset(dataset: Dataset,
                    archive: str,
                    uploader: Optional[ResourceUploader]=None,
                    polling: Union[float, int]=5):
    if not uploader:
        uploader = SyncResourceUploader()
    q = Queue()
    t_dataset = Thread(
        target=__do_upload,
        args=[q],
        kwargs={
            'resource_id': dataset.id,
            'file_path': archive,
            'uploader': uploader,
        }
    )
    logger.info(f"Uploading dataset {dataset.name}...")
    t_dataset.start()
    err = q.get()
    t_dataset.join()
    if err:
        ex_type, ex_value, tb_str = err
        message = f"{str(ex_value)} (in subprocess)\n{tb_str}"
        raise ex_type(message)
    finished = [ResourceState.ERROR, ResourceState.READY]
    while dataset.state not in finished:
        time.sleep(polling)
        dataset.get_detail()
    if dataset.state == ResourceState.ERROR:
        raise AnyLearnException("Error occured when uploading dataset")
    logger.info("Successfully uploaded dataset")


def __do_upload(q: Queue, *args, **kwargs):
    try:
        Resource.upload_file(*args, **kwargs)
        err = None
    except:
        ex_type, ex_value, tb = sys.exc_info()
        err = ex_type, ex_value, ''.join(traceback.format_tb(tb))
    q.put(err)



"""add train hpo id hpo final metric

Revision ID: 6dfd859f129e
Revises: 848203d98d73
Create Date: 2021-05-18 10:54:32.073962

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '6dfd859f129e'
down_revision = '848203d98d73'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('local_train_task', schema=None) as batch_op:
        batch_op.add_column(sa.Column('hpo_final_metric', sa.DECIMAL(), nullable=True))
        batch_op.add_column(sa.Column('hpo_id', sa.String(length=8), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('local_train_task', schema=None) as batch_op:
        batch_op.drop_column('hpo_id')
        batch_op.drop_column('hpo_final_metric')

    # ### end Alembic commands ###



from .algorithm_manager import sync_algorithm
from .tracking import (
    INCONTAINER_TRAIN_TASK_ID,
    INCONTAINER_TRAIN_TASK_SECRET,
    report_intermediate_metric,
    report_final_metric
)
from .quickstart import quick_train
